### USER FILE PARSER DICTIONARY TEMPLATE ###
### DO NOT EDIT THIS FILE DIRECTLY!      ###
### EDIT 'USER_FILE_PARSER.py' INSTEAD!  ###
# -*- coding: utf-8 -*-
"""
plaid - plaid looks at integrated data
F.H. Gjørup 2025-2026
Aarhus University, Denmark
MAX IV Laboratory, Lund University, Sweden

User-defined HDF5 file parser dictionary.
Use this dictionary to define custom dataset paths for loading data from HDF5 files.
Each key corresponds to a specific data type, and the value should be the dataset path as a string.
Only "I" and "tth" or "q" are required. Other fields can be set to None if not available.

Restart the application after editing this file to apply changes.

Example:
USER_FILE_PARSER = {"I": "entry/data/intensity",
                    "tth": "entry/data/two_theta",
                    }
"""

USER_FILE_PARSER = {"I": None,
                    "I_error": None,
                    "tth": None,
                    "q": None,
                    "energy": None,
                    "wavelength": None,
                    "I0": None,
                    "instrument_name": None,
                    "source_name": None,
                    "map_shape": None,
                    "map_indices": None,
                    }